import posixpath

import django
from django.template.exceptions import TemplateDoesNotExist
from django.urls import Resolver404, ResolverMatch, URLPattern
from django.urls.resolvers import RegexPattern

from . import conf
from .helpers import get_template_by_name, safe_join
from .views import serve

if django.VERSION >= (4, 1):  # noqa
    class FlatlyURLPattern(URLPattern):
        def resolve(self, path):
            match = self.pattern.match(path)
            if match:
                new_path, args, captured_kwargs = match
                # Pass any default args as **kwargs.
                kwargs = {**captured_kwargs, **self.default_args}

                path = kwargs.get('path')
                path = posixpath.normpath(path).lstrip('/')
                path = path.replace('-', '_')

                if conf.TEMPLATE_ROOT:
                    template_name = safe_join(path, conf.TEMPLATE_ROOT)
                else:
                    template_name = path

                try:
                    template = get_template_by_name(template_name)
                except TemplateDoesNotExist:
                    raise Resolver404({})

                return ResolverMatch(
                    self.callback,
                    (template,),
                    {},
                    route=str(self.pattern),
                    captured_kwargs=captured_kwargs,
                    extra_kwargs=self.default_args,
                )

else:
    class FlatlyURLPattern(URLPattern):
        def resolve(self, path):
            match = self.pattern.match(path)
            if match:
                new_path, args, kwargs = match

                path = kwargs.get('path')
                path = posixpath.normpath(path).lstrip('/')
                path = path.replace('-', '_')

                if conf.TEMPLATE_ROOT:
                    template_name = safe_join(path, conf.TEMPLATE_ROOT)
                else:
                    template_name = path

                try:
                    template = get_template_by_name(template_name)
                except TemplateDoesNotExist:
                    raise Resolver404({})

                if django.VERSION >= (2, 2):
                    return ResolverMatch(
                        self.callback, (template,), {}, route=str(self.pattern)
                    )
                else:
                    return ResolverMatch(self.callback, (template,), {})


def flatly_path(route, view):
    pattern = RegexPattern(route, name=None, is_endpoint=True)
    return FlatlyURLPattern(pattern, view, None, None)


app_name = 'flatly'
urlpatterns = [
    flatly_path(r'^(?P<path>.*/|)$', serve)
]



from django.conf import settings

TEMPLATE_ROOT = getattr(settings, 'FLATLY_TEMPLATE_ROOT', 'flatly')
ENGINE = getattr(settings, 'FLATLY_ENGINE', None)
EXTENSIONS = getattr(settings, 'FLATLY_EXTENSIONS', ['html'])
CACHE_ENABLED = getattr(settings, 'FLATLY_CACHE_ENABLED', not settings.DEBUG)



import posixpath
from typing import Any, Dict, List

from django.template.base import Origin
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import _engine_list

from . import conf
from .exceptions import InvalidTemplatePath

template_cache = {}  # type: Dict[str, Any]


def safe_join(path, base=None):
    """
    Join path components intelligently.
    Return a normalized version of the template path.
    """
    path = posixpath.normpath(path)
    base = base.strip().strip('/') if base is not None else None

    if base:
        path = posixpath.normpath(posixpath.join(base, path))

        # prevent "../" tricks
        if not path.startswith(base + posixpath.sep) and path != base:
            raise InvalidTemplatePath(path)

    return path


def _check_template_file(engine, name, tried):
    try:
        return engine.get_template(name)
    except TemplateDoesNotExist as e:
        tried.extend(e.tried)
    except IsADirectoryError:
        tried.append((Origin(name, name, engine), 'Source does not exist'))


def check_possible_template_paths(engine, name):
    """
    Search path for "/news/article":
        1) /templates/news/article
        2) /templates/news/article.html
        3) /templates/news/article/index.html
    """
    tried = []  # type: List[Origin]

    # check exact path
    template = _check_template_file(engine, name, tried)
    if template is not None:
        return template

    # try to add an extension
    for ext in conf.EXTENSIONS:
        template_name_with_extension = '.'.join([name, ext])
        template = _check_template_file(engine, template_name_with_extension, tried)
        if template is not None:
            return template

    # try to find "index" file in subfolder
    for ext in conf.EXTENSIONS:
        index_template_name = posixpath.join(name, 'index.{}'.format(ext))
        template = _check_template_file(engine, index_template_name, tried)
        if template is not None:
            return template

    raise TemplateDoesNotExist(name, tried=tried, backend=engine)


def get_template_by_name(name: str):
    if conf.CACHE_ENABLED:
        cached = template_cache.get(name)
        if cached:
            return cached

    chain = []
    engines = _engine_list(conf.ENGINE)
    for engine in engines:
        try:
            template = check_possible_template_paths(engine, name)
        except TemplateDoesNotExist as e:
            chain.append(e)
        else:
            template_cache[name] = template
            return template

    raise TemplateDoesNotExist(name, chain=chain)
